﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CG1.Ex3.Mathematics;

namespace CG1.Ex04.Mathematics
{
	public class Matrix44
	{
		#region Properties

		public double[,] M = new double[4,4];

		#endregion


		#region Construction

		public Matrix44()
		{
		}

		public Matrix44(
			double m00, double m01, double m02, double m03,
			double m10, double m11, double m12, double m13,
			double m20, double m21, double m22, double m23,
			double m30, double m31, double m32, double m33)
		{
			M[0, 0] = m00; M[0, 1] = m01; M[0, 2] = m02; M[0, 3] = m03;
			M[1, 0] = m10; M[1, 1] = m11; M[1, 2] = m12; M[1, 3] = m13;
			M[2, 0] = m20; M[2, 1] = m21; M[2, 2] = m22; M[2, 3] = m23;
			M[3, 0] = m30; M[3, 1] = m31; M[3, 2] = m32; M[3, 3] = m33;
		}

		public void Set(
			double m00, double m01, double m02, double m03,
			double m10, double m11, double m12, double m13,
			double m20, double m21, double m22, double m23,
			double m30, double m31, double m32, double m33)
		{
			M[0, 0] = m00; M[0, 1] = m01; M[0, 2] = m02; M[0, 3] = m03;
			M[1, 0] = m10; M[1, 1] = m11; M[1, 2] = m12; M[1, 3] = m13;
			M[2, 0] = m20; M[2, 1] = m21; M[2, 2] = m22; M[2, 3] = m23;
			M[3, 0] = m30; M[3, 1] = m31; M[3, 2] = m32; M[3, 3] = m33;
		}

		public void Set(Matrix44 m)
		{
			M[0, 0] = m.M[0, 0]; M[0, 1] = m.M[0, 1]; M[0, 2] = m.M[0, 2]; M[0, 3] = m.M[0, 3];
			M[1, 0] = m.M[1, 0]; M[1, 1] = m.M[1, 1]; M[1, 2] = m.M[1, 2]; M[1, 3] = m.M[1, 3];
			M[2, 0] = m.M[2, 0]; M[2, 1] = m.M[2, 1]; M[2, 2] = m.M[2, 2]; M[2, 3] = m.M[2, 3];
			M[3, 0] = m.M[3, 0]; M[3, 1] = m.M[3, 1]; M[3, 2] = m.M[3, 2]; M[3, 3] = m.M[3, 3];
		}

		#endregion


		#region Arithmetic Operations

		public static Matrix44 operator +(Matrix44 a, Matrix44 b)
		{
			Matrix44 result = new Matrix44();

			for (Int32 i = 0; i < 4; i++)
			{
				for (Int32 j = 0; j < 4; j++)
				{
					result.M[i, j] = a.M[i, j] + b.M[i, j];
				}
			}

			return result;
		}

		public static Matrix44 operator -(Matrix44 a, Matrix44 b)
		{
			Matrix44 result = new Matrix44();

			for (Int32 i = 0; i < 4; i++)
			{
				for (Int32 j = 0; j < 4; j++)
				{
					result.M[i, j] = a.M[i, j] - b.M[i, j];
				}
			}

			return result;
		}

		public static Matrix44 operator *(Matrix44 a, Matrix44 b)
		{
			Matrix44 result = new Matrix44();

			for (Int32 i = 0; i < 4; i++)
			{
				for (Int32 j = 0; j < 4; j++)
				{
					double mij = 0;

					for (Int32 k = 0; k < 4; k++)
					{
						mij += a.M[i, k] * b.M[k, j];
					}
					result.M[i, j] = mij;
				}
			}

			return result;
		}

		public static Vector4 operator *(Vector4 a, Matrix44 b)
		{
			Vector4 result = new Vector4();

			result.X = a.X * b.M[0, 0] + a.Y * b.M[1, 0] + a.Z * b.M[2, 0] + a.W * b.M[3, 0];
			result.Y = a.X * b.M[0, 1] + a.Y * b.M[1, 1] + a.Z * b.M[2, 1] + a.W * b.M[3, 1];
			result.Z = a.X * b.M[0, 2] + a.Y * b.M[1, 2] + a.Z * b.M[2, 2] + a.W * b.M[3, 2];
			result.W = a.X * b.M[0, 3] + a.Y * b.M[1, 3] + a.Z * b.M[2, 3] + a.W * b.M[3, 3];

			return result;
		}

		public static Vector4 operator *(Matrix44 a, Vector4 b)
		{
			Vector4 result = new Vector4();

			result.X = a.M[0, 0] * b.X + a.M[0, 1] * b.Y + a.M[0, 2] * b.Z + a.M[0, 3] * b.W;
			result.Y = a.M[1, 0] * b.X + a.M[1, 1] * b.Y + a.M[1, 2] * b.Z + a.M[1, 3] * b.W;
			result.Z = a.M[2, 0] * b.X + a.M[2, 1] * b.Y + a.M[2, 2] * b.Z + a.M[2, 3] * b.W;
			result.W = a.M[3, 0] * b.X + a.M[3, 1] * b.Y + a.M[3, 2] * b.Z + a.M[3, 3] * b.W;

			return result;
		}

		#endregion

		
		#region Transformtions

		public static Matrix44 Zero()
		{
			return new Matrix44(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
		}

		public static Matrix44 Identity()
		{
			return new Matrix44(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);
		}

		public static Matrix44 Scale(Vector4 scaleVector)
		{
			Matrix44 result = Identity();
			result.M[0, 0] = scaleVector.X;
			result.M[1, 1] = scaleVector.Y;
			result.M[2, 2] = scaleVector.Z;

			return result;
		}

		public static Matrix44 Translate(Vector4 translateVector)
		{
			Matrix44 result = Identity();
			result.M[0, 3] = translateVector.X;
			result.M[1, 3] = translateVector.Y;
			result.M[2, 3] = translateVector.Z;

			return result;
		}

		public static Matrix44 RotateX(double angleDeg)
		{
			double angleRad = MathEx.DegToRad(angleDeg);
			Matrix44 result = Identity();
			result.M[1, 1] = Math.Cos(angleRad);
			result.M[1, 2] = -Math.Sin(angleRad);
			result.M[2, 1] = Math.Sin(angleRad);
			result.M[2, 2] = Math.Cos(angleRad);

			return result;
		}

		public static Matrix44 RotateY(double angleDeg)
		{
			double angleRad = MathEx.DegToRad(angleDeg);
			Matrix44 result = Identity();
			result.M[0, 0] = Math.Cos(angleRad);
			result.M[0, 2] = Math.Sin(angleRad);
			result.M[2, 0] = -Math.Sin(angleRad);
			result.M[2, 2] = Math.Cos(angleRad);

			return result;
		}

		public static Matrix44 RotateZ(double angleDeg)
		{
			double angleRad = MathEx.DegToRad(angleDeg);
			Matrix44 result = Identity();
			result.M[0, 0] = Math.Cos(angleRad);
			result.M[0, 1] = -Math.Sin(angleRad);
			result.M[1, 0] = Math.Sin(angleRad);
			result.M[1, 1] = Math.Cos(angleRad);

			return result;
		}

		public static Matrix44 RotateXYZ(Vector4 angleDeg)
		{
			return RotateX(angleDeg.Z) * RotateX(angleDeg.Y) * RotateX(angleDeg.X);
		}

		public Matrix44 Transpose()
		{
			Matrix44 result = new Matrix44();

			for (int i = 0; i < 4; i++)
			{
				for (int j = 0; j < 4; j++)
				{
					result.M[i,j] = M[j,i];
				}
			}

			return result;
		}

		public double Determinant()
		{
			return
			M[0, 3] * M[1, 2] * M[2, 1] * M[3, 0] - M[0, 2] * M[1, 3] * M[2, 1] * M[3, 0] - M[0, 3] * M[1, 1] * M[2, 2] * M[3, 0] + M[0, 1] * M[1, 3] * M[2, 2] * M[3, 0] +
			M[0, 2] * M[1, 1] * M[2, 3] * M[3, 0] - M[0, 1] * M[1, 2] * M[2, 3] * M[3, 0] - M[0, 3] * M[1, 2] * M[2, 0] * M[3, 1] + M[0, 2] * M[1, 3] * M[2, 0] * M[3, 1] +
			M[0, 3] * M[1, 0] * M[2, 2] * M[3, 1] - M[0, 0] * M[1, 3] * M[2, 2] * M[3, 1] - M[0, 2] * M[1, 0] * M[2, 3] * M[3, 1] + M[0, 0] * M[1, 2] * M[2, 3] * M[3, 1] +
			M[0, 3] * M[1, 1] * M[2, 0] * M[3, 2] - M[0, 1] * M[1, 3] * M[2, 0] * M[3, 2] - M[0, 3] * M[1, 0] * M[2, 1] * M[3, 2] + M[0, 0] * M[1, 3] * M[2, 1] * M[3, 2] +
			M[0, 1] * M[1, 0] * M[2, 3] * M[3, 2] - M[0, 0] * M[1, 1] * M[2, 3] * M[3, 2] - M[0, 2] * M[1, 1] * M[2, 0] * M[3, 3] + M[0, 1] * M[1, 2] * M[2, 0] * M[3, 3] +
			M[0, 2] * M[1, 0] * M[2, 1] * M[3, 3] - M[0, 0] * M[1, 2] * M[2, 1] * M[3, 3] - M[0, 1] * M[1, 0] * M[2, 2] * M[3, 3] + M[0, 0] * M[1, 1] * M[2, 2] * M[3, 3];
		}

		public Matrix44 Inverse()
		{
			double idet = Determinant();

			if (idet == 0) return Zero();
			idet = 1 / idet;

			Matrix44 result = new Matrix44();

			result.M[0, 0] = idet * (M[1, 2] * M[2, 3] * M[3, 1] - M[1, 3] * M[2, 2] * M[3, 1] + M[1, 3] * M[2, 1] * M[3, 2] - M[1, 1] * M[2, 3] * M[3, 2] - M[1, 2] * M[2, 1] * M[3, 3] + M[1, 1] * M[2, 2] * M[3, 3]);
			result.M[0, 1] = idet * (M[0, 3] * M[2, 2] * M[3, 1] - M[0, 2] * M[2, 3] * M[3, 1] - M[0, 3] * M[2, 1] * M[3, 2] + M[0, 1] * M[2, 3] * M[3, 2] + M[0, 2] * M[2, 1] * M[3, 3] - M[0, 1] * M[2, 2] * M[3, 3]);
			result.M[0, 2] = idet * (M[0, 2] * M[1, 3] * M[3, 1] - M[0, 3] * M[1, 2] * M[3, 1] + M[0, 3] * M[1, 1] * M[3, 2] - M[0, 1] * M[1, 3] * M[3, 2] - M[0, 2] * M[1, 1] * M[3, 3] + M[0, 1] * M[1, 2] * M[3, 3]);
			result.M[0, 3] = idet * (M[0, 3] * M[1, 2] * M[2, 1] - M[0, 2] * M[1, 3] * M[2, 1] - M[0, 3] * M[1, 1] * M[2, 2] + M[0, 1] * M[1, 3] * M[2, 2] + M[0, 2] * M[1, 1] * M[2, 3] - M[0, 1] * M[1, 2] * M[2, 3]);
			result.M[1, 0] = idet * (M[1, 3] * M[2, 2] * M[3, 0] - M[1, 2] * M[2, 3] * M[3, 0] - M[1, 3] * M[2, 0] * M[3, 2] + M[1, 0] * M[2, 3] * M[3, 2] + M[1, 2] * M[2, 0] * M[3, 3] - M[1, 0] * M[2, 2] * M[3, 3]);
			result.M[1, 1] = idet * (M[0, 2] * M[2, 3] * M[3, 0] - M[0, 3] * M[2, 2] * M[3, 0] + M[0, 3] * M[2, 0] * M[3, 2] - M[0, 0] * M[2, 3] * M[3, 2] - M[0, 2] * M[2, 0] * M[3, 3] + M[0, 0] * M[2, 2] * M[3, 3]);
			result.M[1, 2] = idet * (M[0, 3] * M[1, 2] * M[3, 0] - M[0, 2] * M[1, 3] * M[3, 0] - M[0, 3] * M[1, 0] * M[3, 2] + M[0, 0] * M[1, 3] * M[3, 2] + M[0, 2] * M[1, 0] * M[3, 3] - M[0, 0] * M[1, 2] * M[3, 3]);
			result.M[1, 3] = idet * (M[0, 2] * M[1, 3] * M[2, 0] - M[0, 3] * M[1, 2] * M[2, 0] + M[0, 3] * M[1, 0] * M[2, 2] - M[0, 0] * M[1, 3] * M[2, 2] - M[0, 2] * M[1, 0] * M[2, 3] + M[0, 0] * M[1, 2] * M[2, 3]);
			result.M[2, 0] = idet * (M[1, 1] * M[2, 3] * M[3, 0] - M[1, 3] * M[2, 1] * M[3, 0] + M[1, 3] * M[2, 0] * M[3, 1] - M[1, 0] * M[2, 3] * M[3, 1] - M[1, 1] * M[2, 0] * M[3, 3] + M[1, 0] * M[2, 1] * M[3, 3]);
			result.M[2, 1] = idet * (M[0, 3] * M[2, 1] * M[3, 0] - M[0, 1] * M[2, 3] * M[3, 0] - M[0, 3] * M[2, 0] * M[3, 1] + M[0, 0] * M[2, 3] * M[3, 1] + M[0, 1] * M[2, 0] * M[3, 3] - M[0, 0] * M[2, 1] * M[3, 3]);
			result.M[2, 2] = idet * (M[0, 1] * M[1, 3] * M[3, 0] - M[0, 3] * M[1, 1] * M[3, 0] + M[0, 3] * M[1, 0] * M[3, 1] - M[0, 0] * M[1, 3] * M[3, 1] - M[0, 1] * M[1, 0] * M[3, 3] + M[0, 0] * M[1, 1] * M[3, 3]);
			result.M[2, 3] = idet * (M[0, 3] * M[1, 1] * M[2, 0] - M[0, 1] * M[1, 3] * M[2, 0] - M[0, 3] * M[1, 0] * M[2, 1] + M[0, 0] * M[1, 3] * M[2, 1] + M[0, 1] * M[1, 0] * M[2, 3] - M[0, 0] * M[1, 1] * M[2, 3]);
			result.M[3, 0] = idet * (M[1, 2] * M[2, 1] * M[3, 0] - M[1, 1] * M[2, 2] * M[3, 0] - M[1, 2] * M[2, 0] * M[3, 1] + M[1, 0] * M[2, 2] * M[3, 1] + M[1, 1] * M[2, 0] * M[3, 2] - M[1, 0] * M[2, 1] * M[3, 2]);
			result.M[3, 1] = idet * (M[0, 1] * M[2, 2] * M[3, 0] - M[0, 2] * M[2, 1] * M[3, 0] + M[0, 2] * M[2, 0] * M[3, 1] - M[0, 0] * M[2, 2] * M[3, 1] - M[0, 1] * M[2, 0] * M[3, 2] + M[0, 0] * M[2, 1] * M[3, 2]);
			result.M[3, 2] = idet * (M[0, 2] * M[1, 1] * M[3, 0] - M[0, 1] * M[1, 2] * M[3, 0] - M[0, 2] * M[1, 0] * M[3, 1] + M[0, 0] * M[1, 2] * M[3, 1] + M[0, 1] * M[1, 0] * M[3, 2] - M[0, 0] * M[1, 1] * M[3, 2]);
			result.M[3, 3] = idet * (M[0, 1] * M[1, 2] * M[2, 0] - M[0, 2] * M[1, 1] * M[2, 0] + M[0, 2] * M[1, 0] * M[2, 1] - M[0, 0] * M[1, 2] * M[2, 1] - M[0, 1] * M[1, 0] * M[2, 2] + M[0, 0] * M[1, 1] * M[2, 2]);

			return result;
		}

		#endregion

	}
}
